
import { assert, lowerBound, upperBound, paddedIp } from './utils';

import * as d3 from './d3';
import * as d3Selection from 'd3-selection';
import _ from 'lodash';


// Constants

export const minColDistance = 150;
export const minRefSizeX = 400;

export const padLeft = 60;
export const padRight = padLeft;

const sortIteratees = {
  string: _.identity,
  ip: paddedIp
};


// Local Functions

function numericScale(col, colRange) {
  switch (col.scaleType) {
    case 'linear':
      return d3.scaleLinear()
        .domain(colRange)
        .range([0, 1]);

    case 'log':
      return d3.scaleLog()
        .domain([Math.max(colRange[0], 1), Math.max(colRange[1], 1)])
        .range([0, 1])
        .clamp(true);

    case 'sqrt':
      return d3.scaleSqrt()
        .domain([Math.max(colRange[0], 0), Math.max(colRange[1], 0)])
        .range([0, 1])
        .clamp(true);
  }
}

function makeRangedToText(toText) {
  return function (colType, colRange, sample) {
    colType = colType.slice(0, -6);                   // Remove _range suffix

    const from = (sample.from !== null)
      ? toText(colType, colRange.values, sample.from)
      : '*';

    const to = (sample.to !== null)
      ? toText(colType, colRange.values, sample.to)
      : '*';

    return `[${from}, ${to})`;
  };
}

const rangeToText = makeRangedToText(sampleToText);
const rangeToShortText = makeRangedToText(sampleToShortText);


// Geometry

export function rectContains(rect, pos) {
  return rect[0] <= pos[0] && pos[0] <= rect[2]
    && rect[1] <= pos[1] && pos[1] <= rect[3];
}

export function movedInsideRect(rect, pos) {
  return [
    Math.min(Math.max(pos[0], rect[0]), rect[2]),
    Math.min(Math.max(pos[1], rect[1]), rect[3])
  ];
}

export function translateStep(xLen, len) {
  return len > 1 ? xLen / (len - 1) : 0;
}

export function translator(xLen, len) {
  let step = translateStep(xLen, len);

  return function (d, c) {
    return `translate(${c * step} 0)`;
  };
}

export function closestPoint(ctx, lineToX, lineToY, x, y) {
  let min = Infinity;

  const line = _.reduce(ctx.filtered.records[0], function (memo, line) {
    const dx = lineToX(line) - x;
    const dy = lineToY(line) - y;

    const val = dx * dx + dy * dy;
    return (val <= min) ? (min = val, line) : memo;
  }, null);

  return line ? { line } : {};
}

export function sampleToText(colType, colRange, sample) {
  if (sample === undefined) { return ''; }

  switch (colType) {
    case 'number':
      return colRange[1] <= -1e5 || 1e5 <= colRange[1]
        ? d3.format(',')(_.round(sample, 2))
        : '' + _.round(sample, 2);

    case 'integer':
      return colRange[1] <= -1e5 || 1e5 <= colRange[1]
        ? d3.format(',')(_.round(sample))
        : '' + _.round(sample);

    case 'date':
      return sample.toLocaleString();

    case 'number_range':
    case 'integer_range':
    case 'date_range':
      return rangeToText(colType, colRange, sample);

    default:
      return sample;
  }
}

export function sampleToShortText(colType, colRange, sample) {
  if (sample === undefined) { return ''; }

  switch (colType) {
    case 'string':
      return sample.slice(0, 3);

    case 'date':
      return sample.toLocaleDateString();

    case 'date_range':
      return rangeToShortText(colType, colRange, sample);

    default:
      return sampleToText(colType, colRange, sample);
  }
}

export function lowerSampleInValueRange(colType, scale, v0, v1) {
  if (scale.invert) { return scale.invert(v0); }

  const domain = scale.domain();
  const lo = (v0 <= 0)
    ? 0
    : lowerBound(scale.domain(), v0, sample => scale(sample), sortIteratees[colType]);

  return lo < domain.length && scale(domain[lo]) <= v1 ? domain[lo] : '';
}

export function upperSampleInValueRange(colType, scale, v0, v1) {
  if (scale.invert) { return scale.invert(v1); }

  const domain = scale.domain();
  const hi = (v1 >= 1)
    ? domain.length - 1
    : upperBound(scale.domain(), v1, sample => scale(sample), sortIteratees[colType]);

  return hi >= 0 && v0 <= scale(domain[hi]) ? domain[hi] : '';
}


// Interaction

export function rawMousePos(ctx, event) {
  // Rationale: mouse hover events don't have an event in d3.event,
  // so we have to supply it manually.

  // See https://brianschiller.com/blog/2017/09/28/d3-event-issues
  // I tried customEvent, too - but couldn't make it work; it should
  // probably implemented that way however. If this breaks in the
  // future, look into it.

  const currEvent = d3Selection.event;

  d3Selection.event = event;
  const result = d3.mouse(d3.select(ctx.svg).node());
  d3Selection.event = currEvent;

  return result;
}

export function mousePos(ctx, event) {
  const rawPos = rawMousePos(ctx, event);
  return rawPos && movedInsideRect(ctx.contentRect, rawPos);
}

export function valueFilter(ctx, c, sample) {
  const v = ctx.scales.columns[c](sample);

  return {
    type: 'value',
    v0: v,
    v1: v,
    s1: sample,
    highlighted: false
  };
}

export function rangeFilter(ctx, c, v0, v1) {
  const col = ctx.columns[c];
  const scale = ctx.scales.columns[c];

  v0 = Math.min(Math.max(v0, 0), 1);
  v1 = Math.min(Math.max(v1, 0), 1);

  return {
    type: 'range',
    v0,
    v1,
    s0: lowerSampleInValueRange(col.type, scale, v0, v1),
    s1: upperSampleInValueRange(col.type, scale, v0, v1)
  };
}

export function filterDefineEnd(ctx) {
  ctx.mousePress = null;

  ctx.view.colFilters = _.transform(ctx.view.colFilters, function (memo, filter, key) {
    if (filter.type !== 'range' || filter.v0 < filter.v1) { memo[key] = filter; }
  }, {});

  ctx.viewChanges.colFilters = true;

  refresh(ctx);
}

export function resetHover(ctx) {
  if (!ctx.view.hover.line) { return; }

  ctx.view.hover = {};
  ctx.viewChanges.hover = true;

  refresh(ctx);
}


// Rendering

export function materialize(parent, appendName, className, data, key) {
  let selector = appendName;
  if (className) { selector += '.' + className.replace(' ', '.'); }

  let result = parent.selectAll(selector).data(data, key);
  result.exit().remove();

  return result.enter().append(appendName)
    .attr('class', className)
    .merge(result);
}

export function columnTicks(column, scale, columnSize, ticksDistance) {
  const domain = scale.domain();

  ticksDistance = Math.max(20, ticksDistance || 0);
  let ticksCount = Math.max(Math.floor(columnSize / ticksDistance), 3);

  switch (column.type) {
    case 'string':
    case 'ip':
    case 'number_range':
    case 'integer_range':
    case 'date_range':
      ticksCount = Math.min(ticksCount, domain.length - 1);
      return _.map(d3.ticks(0, domain.length - 1, ticksCount),
        s => domain[s]);

    case 'integer':
      ticksCount = Math.min(ticksCount, domain[1] - domain[0] + 1);
      // continue
  }

  if (column.scaleType !== 'log') {
    return scale.ticks(ticksCount);
  } else {
    const magnitudesCount = 3;

    // See d3-scale docs, log scale ticks are magnitude-based...
    const ticks = scale.ticks(magnitudesCount);
    const tickFormat = scale.tickFormat(ticksCount);

    return ticks.map(tickFormat);
  }
}

export function refresh(ctx) {
  if (_.isEmpty(ctx.viewChanges)) { return; }

  ctx.onViewChanges(ctx.viewChanges);
  ctx.viewChanges = {};
}


// State

export function scalesFromRenderData(renderData) {
  const { columns, colRanges } = renderData;

  const columnScales = _.map(columns, (col, c) => {
    const colRange = colRanges[c];

    switch (col.type) {
      case 'number':
      case 'integer':
        return numericScale(col, colRange);

      case 'date':
        return d3.scaleTime()
          .domain(colRange)
          .range([0, 1]);

      case 'string':
      case 'ip':
        return d3.scalePoint()
          .domain(colRange)
          .range([0, 1]);

      case 'number_range':
      case 'integer_range':
      case 'date_range':
        const rawScale = numericScale(col, colRange.values);
        const scale = _.flow(_.iteratee('value'), rawScale);

        _.assign(scale, _.pick(rawScale, [ 'range', 'ticks', 'tickFormat' ]));
        scale.domain = () => colRange.samples;

        return scale;

      default:
        assert(false, `Unrecognized column type '${col.type}'`);
    }
  });


  const colorRange = renderData.colorRange;
  const colors = renderData.colors || {};
  const lowCountColor = colors.low || 'black';
  const highCountColor = colors.high || 'black';


  const colorScale = d3
    .scaleSequential(d3.interpolateRgb(lowCountColor, highCountColor))
    .domain(colorRange);

  return { columns: columnScales, color: colorScale };
}

export function buildStripSizes(ctx, padding) {
  let result = {};

  result.refSizeX = Math.max((ctx.columns.length - 1) * minColDistance, minRefSizeX);
  result.refSizeY = Math.max(ctx.sizeY - padding.top - padding.bottom, 1);

  let surplusPaddingX = Math.max(
    (ctx.sizeX - result.refSizeX - padding.left - padding.right) / 2,
    0);

  result.paddingLeft = padding.left + surplusPaddingX;
  result.paddingRight = padding.right + surplusPaddingX;

  result.sizeX = result.refSizeX + result.paddingLeft + result.paddingRight;

  result.viewBox = [
    -result.paddingLeft, -padding.top,
    result.sizeX, ctx.sizeY
  ];

  result.viewRect = [
    -result.paddingLeft, -padding.top,
    result.sizeX - result.paddingLeft, ctx.sizeY - padding.top
  ];

  result.contentRect = [ 0, 0, result.refSizeX, result.refSizeY ];

  return result;
}

