
import * as cc from './chartcommons';

import * as d3 from './d3';
import _ from 'lodash';

import '../styles/parallel_chart.css';


// Consts

const paddingY = 0;
const textLineHeight = 24;
const fontSize = textLineHeight / 2;
const fontUpShift = .75 * fontSize;
const topToolbarHeight = 2 * textLineHeight;
const botToolbarHeight = textLineHeight;
const padTop = paddingY + topToolbarHeight;
const padBottom = paddingY + botToolbarHeight;

const splineCP = .5;
const borderAxisSplit = 2;

const mouseFilterTolerance = 10;
const filterRangeWidth = 20;
const filterRangeWidthHalf = filterRangeWidth / 2;
const filterRangeHandlesHeight = 5;
const filterValueRadius = 5;
const filterIconsPad = 15;

const labelPositionsByFilter = {
  none: { top: [fontUpShift], bottom: [fontUpShift] },
  range: { top: [fontUpShift], bottom: [fontUpShift] },
  value: { top: [fontUpShift + filterValueRadius], bottom: [] }
};

const sortIcons = { none: '\uf0dc', asc: '\uf0de', desc: '\uf0dd' };


// Local Functions

// Geometry

function xToColumnIndex(ctx, x) {
  return Math.floor(ctx.xScale.invert(x));
}

function xToColumnPos(ctx, c, x) {
  let xScale = ctx.xScale;

  let x0 = xScale(c);
  let x1 = xScale(c + 1);

  return (x - x0) / (x1 - x0);
}

function xToClosestColumnIndex(ctx, x) {
  const c = xToColumnIndex(ctx, x);
  return xToColumnPos(ctx, c, x) <= .5 ? c : c + 1;
}

function valueToY(ctx, val) {
  // Y scales inverted due to svg y axis top->bot
  return ctx.refSizeY * (1 - val);
}

function yToValue(ctx, y) {
  return 1 - y / ctx.refSizeY;
}


function cubicX(x, t) {
  /* eslint-disable  camelcase, space-unary-ops */
  let one_t = 1 - t;

  let c1 = splineCP;
  let c2 = 1 - splineCP;

  let mid = (one_t * c1 + t * c2);

  return (one_t * (one_t * (/*one_t * 0*/ + t * c1) + t * mid)
    + t * (one_t * mid + t * (one_t * c2 + t /** 1*/)))
    - x;
  /* eslint-enable  camelcase, space-unary-ops */
}

function cubicXDerivative(t) {
  /* eslint-disable  camelcase */
  let one_t = 1 - t;

  let c1 = splineCP;
  let c2 = 1 - splineCP;
  let c2_c1 = c2 - c1;

  return 3 * (one_t * (one_t * (c1 /*- 0*/) + t * c2_c1) +
    t * (one_t * c2_c1 + t * (1 - c2)));
  /* eslint-enable  camelcase */
}

function newtonCubicX(x, t) {
  return t - cubicX(x, t) / cubicXDerivative(t);
}

function closestSpline(ctx, posX, posY) {
  const c = Math.min(xToColumnIndex(ctx, posX), ctx.columns.length - 2);
  if (c < 0) { return {}; }

  let x = xToColumnPos(ctx, c, posX);
  let t = x;

  t = newtonCubicX(x, t);
  t = newtonCubicX(x, t);
  t = newtonCubicX(x, t);

  /* eslint-disable  camelcase */
  const one_t = 1 - t;

  const yScale0 = ctx.yScales[c];
  const yScale1 = ctx.yScales[c + 1];

  const l0 = ctx.visibleToAll[c];
  const l1 = ctx.visibleToAll[c + 1];

  let min = Infinity;
  const line = _.reduce(ctx.filtered.records[0], function (memo, line) {
    const y0 = yScale0(line[l0]);
    const y1 = yScale1(line[l1]);

    const y0_1 = one_t * y0 + t * y1;
    const y = one_t * (one_t * y0 + t * y0_1) + t * (one_t * y0_1 + t * y1);

    const val =  Math.abs(y - posY);
    return (val <= min) ? (min = val, line) : memo;
  }, null);

  return line ? { line } : {};
  /* eslint-enable  camelcase */
}


// Interaction

function updateHover(ctx) {
  let mPos = cc.rawMousePos(ctx, d3.event);
  if (!cc.rectContains(ctx.hoverRect, mPos)) { return cc.resetHover(ctx); }

  mPos = cc.movedInsideRect(ctx.contentRect, mPos);

  const hover = closestSpline(ctx, mPos[0], mPos[1]);
  if (hover.line === ctx.view.hover.line) { return; }

  hover.event = d3.event;

  ctx.view.hover = hover;
  ctx.viewChanges.hover = true;

  cc.refresh(ctx);
}


function filterDefineFilter(ctx) {
  const mPos = d3.mouse(ctx.svg);

  if (d3.event.button || !cc.rectContains(ctx.hoverRect, mPos)) {
    return false;
  }

  const c = xToClosestColumnIndex(ctx, mPos[0]);
  const dx = mPos[0] - ctx.xScale(c);

  return Math.abs(dx) < filterRangeWidthHalf;
}

function filterDefineStart(ctx) {
  const mPos = [d3.event.x, d3.event.y];

  ctx.mousePress = {
    mPos,
    c: xToClosestColumnIndex(ctx, mPos[0]),
    v: yToValue(ctx, mPos[1])
  };

  toggleValueFilter(ctx);
}

function toggleValueFilter(ctx) {
  const line = ctx.view.hover.line;
  if (!line) { return; }

  ctx.viewChanges.colFilters = true;

  const c = ctx.mousePress.c;
  const colId = ctx.columns[c].id;

  const filter = ctx.view.colFilters[colId];
  if (filter && filter.type === 'value') {
    delete ctx.view.colFilters[colId];
    return cc.refresh(ctx);
  }

  const l = ctx.visibleToAll[c];

  ctx.view.colFilters[colId] = cc.valueFilter(ctx, c, line[l]);

  cc.resetHover(ctx);
}

function filterDefineMove(ctx) {
  const mPos = [d3.event.x, d3.event.y];

  const c = ctx.mousePress.c;
  const colId = ctx.columns[c].id;
  const filter = ctx.view.colFilters[colId];

  if (!(filter && filter.type === 'range') &&
      Math.abs(mPos[1] - ctx.mousePress.mPos[1]) < mouseFilterTolerance) {

    return;
  }

  const v = yToValue(ctx, mPos[1]);
  const v0 = Math.min(v, ctx.mousePress.v);
  const v1 = Math.max(v, ctx.mousePress.v);

  ctx.view.colFilters[colId] = cc.rangeFilter(ctx, c, v0, v1);
  ctx.viewChanges.colFilters = true;

  cc.refresh(ctx);
}

function filterRemove(ctx, c) {
  d3.event.stopPropagation();

  delete ctx.view.colFilters[ctx.columns[c].id];

  ctx.viewChanges.colFilters = true;
  cc.refresh(ctx);
}

function filterDragStart(ctx, c) {
  const filter = ctx.view.colFilters[ctx.columns[c].id];

  ctx.mousePress = {
    c,
    v0: filter.v0,
    v1: filter.v1,
    vPress: yToValue(ctx, d3.event.y)
  };

  cc.resetHover(ctx);
  cc.refresh(ctx);
}

function filterDragMove(ctx) {
  const { c, v0, v1, vPress } = ctx.mousePress;
  const drag = yToValue(ctx, d3.event.y) - vPress;
  const colId = ctx.columns[c].id;

  ctx.view.colFilters[colId] = cc.rangeFilter(ctx, c, v0 + drag, v1 + drag);
  ctx.viewChanges.colFilters = true;

  cc.refresh(ctx);
}

function filterTopStart(ctx, c) {
  const filter = ctx.view.colFilters[ctx.columns[c].id];

  ctx.mousePress = {
    c,
    v0: filter.v0,
    v1: filter.v1,
    vPress: yToValue(ctx, d3.event.y)
  };

  cc.resetHover(ctx);
}

function filterTopMove(ctx) {
  const { c, v0, v1, vPress } = ctx.mousePress;
  const drag = yToValue(ctx, d3.event.y) - vPress;
  const colId = ctx.columns[c].id;

  ctx.view.colFilters[colId] = cc.rangeFilter(
      ctx, c, v0, Math.max(v0, v1 + drag));

  ctx.viewChanges.colFilters = true;
  cc.refresh(ctx);
}


function filterBotStart(ctx, c) {
  const filter = ctx.view.colFilters[ctx.columns[c].id];

  ctx.mousePress = {
    c,
    v0: filter.v0,
    v1: filter.v1,
    vPress: yToValue(ctx, d3.event.y)
  };

  cc.resetHover(ctx);
}

function filterBotMove(ctx) {
  const { c, v0, v1, vPress } = ctx.mousePress;
  const drag = yToValue(ctx, d3.event.y) - vPress;
  const colId = ctx.columns[c].id;

  ctx.view.colFilters[colId] = cc.rangeFilter(
    ctx, c, Math.min(v0 + drag, v1), v1);

  ctx.viewChanges.colFilters = true;
  cc.refresh(ctx);
}


function colDragFilter(ctx) {
  return (d3.event.button === 0) && !ctx.colDrag;
}

function colDragStart(ctx, c) {
  ctx.colDrag = { dstIdx: c, columns: ctx.columns };
  renderOverlayGroup(ctx);
}

function colDragMove(ctx, c) {
  let dstIdx = xToColumnIndex(ctx, d3.event.x);
  let t = xToColumnPos(ctx, dstIdx, d3.event.x);

  if (t > .5) { ++dstIdx; }
  dstIdx = Math.min(Math.max(dstIdx, 0), ctx.columns.length);

  if (dstIdx === ctx.colDrag.dstIdx) { return; }

  let columns = ctx.columns.slice();
  let col = columns[c];

  columns.splice(c, 1);
  columns.splice(dstIdx, 0, col);

  ctx.colDrag.dstIdx = dstIdx;
  ctx.colDrag.columns = columns;

  renderOverlayGroup(ctx);
}

function colDragEnd(ctx, c) {
  let dstIdx = ctx.colDrag.dstIdx;

  if (dstIdx === c) {
    ctx.colDrag = null;
    return renderOverlayGroup(ctx);
  }

  ctx.onMoveColumn(c, dstIdx);
}

function initMouse(ctx) {
  ctx.filterDefineEnd = () => cc.filterDefineEnd(ctx);

  ctx.filterDragStart = c => filterDragStart(ctx, c);
  ctx.filterTopStart = c => filterTopStart(ctx, c);
  ctx.filterBotStart = c => filterBotStart(ctx, c);

  ctx.filterDragMove = () => filterDragMove(ctx);
  ctx.filterTopMove = () => filterTopMove(ctx);
  ctx.filterBotMove = () => filterBotMove(ctx);

  ctx.colDragFilter = c => colDragFilter(ctx, c);
  ctx.colDragStart = c => colDragStart(ctx, c);
  ctx.colDragMove = c => colDragMove(ctx, c);
  ctx.colDragEnd = c => colDragEnd(ctx, c);

  d3.select(ctx.svg)
    .on('mousemove', () => updateHover(ctx))
    .on('mouseleave', () => cc.resetHover(ctx))
    .call(d3.drag()
      .container(ctx.svg)
      .filter(() => filterDefineFilter(ctx))
      .on('start', () => filterDefineStart(ctx))
      .on('drag', () => filterDefineMove(ctx))
      .on('end', ctx.filterDefineEnd));
}


// Rendering

function renderColumnHeader(ctx, group, column, c) {
  let txt = cc.materialize(group, 'text', 'plg-toolbar-col-text', [0])
    .attr('y', -1.25 * textLineHeight);

  cc.materialize(txt, 'tspan', 'plg-toolbar-icon plg-toolbar-left', [0])
    .html('\uf0d9 &thinsp;');                     // caret left

  cc.materialize(txt, 'tspan', 'plg-toolbar-icon plg-toolbar-sort', [c])
    .html(sortIcons[column.sort] + ' &thinsp;')   // sort
    .on('click', ctx.onSortColumn);

  cc.materialize(txt, 'tspan', 'plg-toolbar-col-label', [0])
    .text(column.label);

  cc.materialize(txt, 'tspan', 'plg-toolbar-icon plg-toolbar-hide', [c])
    .html('&thinsp; \uf00d')                      // cross
    .on('click', ctx.onHideColumn);

  cc.materialize(txt, 'tspan', 'plg-toolbar-icon plg-toolbar-right', [0])
    .html('&thinsp; \uf0da');                     // caret left
}

function renderTopToolbar(ctx, group) {
  let colDragging = !!ctx.colDrag;
  let columns = colDragging ? ctx.colDrag.columns : ctx.columns;

  group.classed('plg-dragging', colDragging);

  cc.materialize(group, 'g', 'plg-top-toolbar-column', columns)
    .classed('plg-dragged', (col, c) => colDragging && c === ctx.colDrag.dstIdx)
    .attr('transform', cc.translator(ctx.refSizeX, columns.length))
    .each(function (column, c) { renderColumnHeader(ctx, d3.select(this), column, c); })
    .datum((col, c) => c)
    .call(d3.drag()
      .container(ctx.svg)
      .filter(ctx.colDragFilter)
      .on('start', ctx.colDragStart)
      .on('drag', ctx.colDragMove)
      .on('end', ctx.colDragEnd));
}


function renderAxis(ctx, group) {
  const columns = ctx.columns;
  const colFilters = ctx.view.colFilters;

  cc.materialize(group, 'g', 'plg-column', columns)
    .attr('transform', cc.translator(ctx.refSizeX, columns.length))
    .each(function (column, c) {
      const columnGroup = d3.select(this);

      cc.materialize(columnGroup, 'line', 'plg-column-line', [0])
        .attr('x1', _.identity).attr('x2', _.identity)
        .attr('y1', 0).attr('y2', ctx.refSizeY);


      // Min-Max labels

      const colRange = ctx.colRanges[c];
      const filter = colFilters[column.id];
      const scale = ctx.scales.columns[c];

      let filterType, v0, v1, s0, s1;       // eslint-disable-line  one-var
      if (filter) {
        filterType = filter.type;
        v0 = filter.v0;
        v1 = filter.v1;
        s0 = filter.s0;
        s1 = filter.s1;
      } else {
        filterType = 'none';
        v0 = 0;
        v1 = 1;
        s0 = cc.lowerSampleInValueRange(column.type, scale, v0, v1);
        s1 = cc.upperSampleInValueRange(column.type, scale, v0, v1);
      }

      cc.materialize(columnGroup, 'text', 'plg-column-top-text',
          labelPositionsByFilter[filterType].top)
        .attr('x', 0)
        .attr('y', lPos => valueToY(ctx, v1) - lPos)
        .attr('font-size', fontSize)
        .text(cc.sampleToShortText(column.type, colRange, s1));

      cc.materialize(columnGroup, 'text', 'plg-column-bot-text',
          labelPositionsByFilter[filterType].bottom)
        .attr('x', 0)
        .attr('y', lPos => valueToY(ctx, v0) + fontUpShift + lPos)
        .attr('font-size', fontSize)
        .text(cc.sampleToShortText(column.type, colRange, s0));


      // Custom ticks

      const yScale = ctx.yScales[c];
      const ticks = filter
        ? []
        : cc.columnTicks(column, scale, ctx.refSizeY, ctx.ticksDistance);

      if (ticks[0] === s0) { ticks.shift(); }
      if (ticks[ticks.length - 1] === s1) { ticks.pop(); }

      cc.materialize(columnGroup, 'line', 'plg-column-tick-line', ticks)
        .attr('x1', 0).attr('x2', -4)
        .attr('y1', yScale).attr('y2', yScale);

      cc.materialize(columnGroup, 'text', 'plg-column-tick-text', ticks)
        .attr('x', -5)
        .attr('y', yScale)
        .attr('dy', '0.25em')
        .text(s => cc.sampleToShortText(column.type, colRange, s));
    });
}


function splineGeometry(ctx, cStart, cEnd) {
  const xScale = ctx.xScale;
  const yScales = ctx.yScales;

  return function d(line) {
    if (!line) { return ''; }

    let y0;
    let y1 = yScales[cStart](line[ctx.visibleToAll[cStart]]);
    let res = `M ${xScale(cStart)},${y1} C`;

    for (let c = cStart + 1; c < cEnd; ++c) {
      y0 = y1;
      y1 = yScales[c](line[ctx.visibleToAll[c]]);

      res += ` ${xScale(c - 1 + splineCP)},${y0} ${xScale(c - splineCP)},${y1} ${xScale(c)},${y1}`;
    }

    return res;
  };
}

function renderSplinesSvg(ctx, group) {
  const filteredLines = ctx.filtered.records;
  const cCount = ctx.columns.length;

  if (cCount < 2) { return; }

  const outGroup = cc.materialize(group, 'g', 'plg-filtered-out', [0]);
  const inGroup = cc.materialize(group, 'g', 'plg-filtered-in', [0]);

  const filteredOutLines = ctx.showFilteredOutRecords ? filteredLines[1] : [];
  const geometry = splineGeometry(ctx, 0, cCount);

  cc.materialize(outGroup, 'path', '', filteredOutLines)
    .attr('d', geometry);

  cc.materialize(inGroup, 'path', '', filteredLines[0])
    .attr('d', geometry)
    .attr('stroke', line => ctx.scales.color(line[line.length - 1]));
}

function drawSplineFactory(ctx, canvas, cStart, cEnd) {
  const xScale = ctx.xScale;
  const yScales = ctx.yScales;

  return function draw(line) {
    let y0;
    let y1 = yScales[cStart](line[ctx.visibleToAll[cStart]]);

    canvas.beginPath();
    canvas.moveTo(xScale(cStart), y1);

    for (let c = cStart + 1; c < cEnd; ++c) {
      y0 = y1;
      y1 = yScales[c](line[ctx.visibleToAll[c]]);

      canvas.bezierCurveTo(
        xScale(c - 1 + splineCP), y0,
        xScale(c - splineCP), y1,
        xScale(c), y1);
    }

    canvas.stroke();
  };
}

function renderSplinesCanvas(ctx) {
  const canvas = ctx.canvas.getContext('2d');

  canvas.setTransform(1, 0, 0, 1, 0, 0);
  canvas.clearRect(0, 0, ctx.sizeX, ctx.sizeY);

  canvas.translate(-ctx.viewBox[0], -ctx.viewBox[1]);
  canvas.lineWidth = 1;

  const filteredLines = ctx.filtered.records;
  const cCount = ctx.columns.length;

  if (cCount < 2) { return; }

  const colorScale = ctx.scales.color;
  const drawSpline = drawSplineFactory(ctx, canvas, 0, cCount);

  if (ctx.showFilteredOutRecords) {
    canvas.strokeStyle = 'rgb(230,230,230)';
    _.forEach(filteredLines[1], drawSpline);
  }

  _.forEach(filteredLines[0], function drawColoredSpline(line) {
    canvas.strokeStyle = colorScale(line[line.length - 1]);
    drawSpline(line);
  });
}

function renderHover(ctx, group) {
  const hover = ctx.view.hover;
  const hoverLine = hover.line;

  if (!hoverLine) {
    group.selectAll('.plg-line-hover').remove();
    group.selectAll('.plg-line-hover-txcount').remove();
    group.selectAll('.plg-line-hover-text').remove();

    return;
  }

  const textDisplaceUp = -1.25 * fontSize;
  const textDisplaceDn = 2 * fontSize;
  const refSizeYHalf = ctx.refSizeY / 2;

  const columns = ctx.columns;
  const xScale = ctx.xScale;
  const cCount = ctx.columns.length;


  cc.materialize(group, 'path', 'plg-line-hover', [hoverLine])
    .attr('d', splineGeometry(ctx, 0, cCount))
    .attr('stroke', ctx.colors.highlight || 'black');


  const txtPos = [cc.mousePos(ctx, hover.event)];
  const count = hoverLine[hoverLine.length - 1];

  // NOTE: Text displacements are inverted for bucket count labels, to avoid
  //       sharing space with column value labels

  cc.materialize(group, 'text', 'plg-line-hover-txcount', txtPos)
    .attr('x', m => m[0])
    .attr('y', m =>
      m[1] + (m[1] < refSizeYHalf ? textDisplaceUp : textDisplaceDn))
    .attr('font-size', fontSize)
    .attr('filter', 'url(#text-background)')
    .text(`Count: ${count}`);


  const lineIndexes = _(0).range(cCount)
    .map(c => ({ c, l: ctx.visibleToAll[c] }))
    .value();

  cc.materialize(group, 'text', 'plg-line-hover-text', lineIndexes)
    .attr('x', d => xScale(d.c))
    .attr('y', d => {
      const y = ctx.yScales[d.c](hoverLine[d.l]);
      return y + (y < refSizeYHalf ? textDisplaceDn : textDisplaceUp);
    })
    .attr('font-size', fontSize)
    .attr('filter', 'url(#text-background)')
    .text(d => cc.sampleToText(
      ctx.columns[d.c].type, ctx.colRanges[d.c], hoverLine[d.l]));
}

function renderFilters(ctx, group) {
  const xScale = ctx.xScale;
  const columns = ctx.columns;
  const colFilters = ctx.view.colFilters;

  const colIndices = _(columns.length).range()
    .filter(c => colFilters[columns[c].id])
    .value();

  cc.materialize(group, 'g', 'plg-filter', colIndices)
    .each(function (c) {
      const filterGroup = d3.select(this);
      const filter = ctx.view.colFilters[columns[c].id];

      const x = xScale(c);
      const y0 = valueToY(ctx, filter.v1);    // Scale inversion
      const y1 = valueToY(ctx, filter.v0);

      renderFilterRange(ctx, filterGroup, filter, c, x, y0, y1);
      renderFilterValue(ctx, filterGroup, filter, c, x, y0);
    });
}

function renderFilterRange(ctx, group, filter, c, x, y0, y1) {
  const items = (filter.type === 'range') ? [c] : [];

  const x0 = x - filterRangeWidthHalf;
  const height = y1 - y0;
  const handlesHeight = Math.min(height, filterRangeHandlesHeight);


  cc.materialize(group, 'rect', 'plg-filter-range', items)
    .attr('x', x0).attr('width', filterRangeWidth)
    .attr('y', y0).attr('height', height)
    .call(d3.drag()
      .container(ctx.svg)
      .on('start', ctx.filterDragStart)
      .on('drag', ctx.filterDragMove)
      .on('end', ctx.filterDefineEnd));

  cc.materialize(group, 'rect', 'plg-filter-range-top', items)
    .attr('x', x0).attr('width', filterRangeWidth)
    .attr('y', y0).attr('height', handlesHeight)
    .call(d3.drag()
      .container(ctx.svg)
      .on('start', ctx.filterTopStart)
      .on('drag', ctx.filterTopMove)
      .on('end', ctx.filterDefineEnd));

  cc.materialize(group, 'rect', 'plg-filter-range-bot', items)
    .attr('x', x0)
    .attr('y', Math.max(y1 - filterRangeHandlesHeight, y0))
    .attr('width', filterRangeWidth)
    .attr('height', handlesHeight)
    .call(d3.drag()
      .container(ctx.svg)
      .on('start', ctx.filterBotStart)
      .on('drag', ctx.filterBotMove)
      .on('end', ctx.filterDefineEnd));

  cc.materialize(group, 'text', 'plg-filter-delete', items)
    .attr('x', x + filterRangeWidthHalf + filterIconsPad)
    .attr('y', y0 + filterIconsPad)
    .text('\uf00d')
    .on('mousedown', () => d3.event.stopPropagation())
    .on('mouseup', filter => filterRemove(ctx, c));
}

function renderFilterValue(ctx, group, filter, c, x, y) {
  const filters = (filter.type === 'value') ? [filter] : [];

  cc.materialize(group, 'circle', 'plg-filter-value', filters)
    .attr('cx', x)
    .attr('cy', y)
    .attr('r', flt => flt.highlighted ? 1.5 * filterValueRadius : filterValueRadius)
    .on('mouseover', flt => flt.highlighted = true)
    .on('mouseout', flt => flt.highlighted = false);
}

function renderDefsGroup(ctx) {
  let svg = d3.select(ctx.svg);
  let tipColor = ctx.colors.tip || 'black';

  const filtersGroup = cc.materialize(svg, 'defs', 'plg-defs', [0]);

  const filter = cc.materialize(filtersGroup, 'filter', '', [0])
    .attr('id', 'text-background');

  cc.materialize(filter, 'feFlood', '', [0])
    .attr('flood-color', tipColor)
    .attr('flood-opacity', 0.5);

  cc.materialize(filter, 'feComposite', '', [0])
    .attr('operator', 'over')
    .attr('in', 'SourceGraphic');
}


function renderContentGroup(ctx) {
  let svg = d3.select(ctx.svg);

  let contentGroup = cc.materialize(svg, 'g', 'plg-content', [0]);

  if (ctx.useCanvas) {
    contentGroup.selectAll('*').remove();
    renderSplinesCanvas(ctx);
  } else {
    renderSplinesSvg(ctx, contentGroup);
  }
}

function renderOverlayGroup(ctx) {
  let svg = d3.select(ctx.svg);
  let overlayGroup = cc.materialize(svg, 'g', 'plg-overlay', [0]);

  let topToolbar = cc.materialize(overlayGroup, 'g', 'plg-top-toolbar', [0]);
  renderTopToolbar(ctx, topToolbar);

  let axisGroup = cc.materialize(overlayGroup, 'g', 'plg-axis-group', [0]);
  renderAxis(ctx, axisGroup);

  let hoverGroup = cc.materialize(overlayGroup, 'g', 'plg-hover-group', [0]);
  renderHover(ctx, hoverGroup);

  let filtersGroup = cc.materialize(overlayGroup, 'g', 'plg-filters-group', [0]);
  renderFilters(ctx, filtersGroup);
}


// State

function buildSizes(ctx) {
  let padding = {
    left: cc.padLeft, right: cc.padRight,
    top: padTop, bottom: padBottom
  };

  let result = cc.buildStripSizes(ctx, padding);

  result.hoverRect = [
    result.contentRect[0] - filterRangeWidthHalf,
    -textLineHeight,
    result.contentRect[2] + filterRangeWidthHalf,
    result.viewRect[3]
  ];

  return result;
}

function buildScales(ctx) {
  let result = {};

  result.xScale = d3.scaleLinear()
    .domain([0, ctx.columns.length - 1])
    .range([0, ctx.refSizeX]);

  result.yScales = _.map(ctx.scales.columns, scale => function (val) {
    return valueToY(ctx, scale(val));
  });

  return result;
}

function initState(base, ctx) {
  ctx = _.assign({
    viewChanges: {},
    mousePress: null,
    colDrag: null
  }, base, ctx);

  _.assign(ctx, buildSizes(ctx));
  _.assign(ctx, buildScales(ctx));

  return ctx;
}


// Exports

export default class ParallelChart {
  constructor(opts) {
    this.baseCtx = opts;
    this.ctx = null;
  }

  render(ctx) {
    this.ctx = ctx = initState(this.baseCtx, ctx);

    d3.select(ctx.svg)
      .attr('width', ctx.sizeX)
      .attr('height', ctx.sizeY)
      .attr('viewBox', ctx.viewBox.join(' '));

    ctx.canvas.width = ctx.sizeX;
    ctx.canvas.height = ctx.sizeY;
    ctx.canvas.style.width = ctx.sizeX + 'px';
    ctx.canvas.style.height = ctx.sizeY + 'px';

    initMouse(ctx);

    renderDefsGroup(ctx);
    renderContentGroup(ctx);
    renderOverlayGroup(ctx);
  }

  refresh(viewChanges) {
    if (!this.ctx) { return; }

    let refreshContent = false;
    let refreshOverlay = false;

    _.chain(viewChanges).keys().forEach(viewData => {
      switch (viewData) {
        case 'colFilters':
          refreshContent = true;

        case 'hover':
          refreshOverlay = true;
      }
    }).commit();

    if (refreshContent) { renderContentGroup(this.ctx); }
    if (refreshOverlay) { renderOverlayGroup(this.ctx); }
  }

  reset() {
    if (!this.ctx) { return; }

    d3.select(this.ctx.svg).selectAll('*').remove();
    this.ctx = null;
  }
};

