
import { scalesFromRenderData } from './chartcommons';
import { assert, arrayMoveValue, paddedIp } from './utils';

import _ from 'lodash';


function fieldType(field) {
  if (!field) { return ''; }

  let type;

  switch (field.esType) {
    case 'byte': case 'short': case 'integer': case 'long':
      type = 'integer';
      break;

    default:
      type = field.type;
  }

  return type;
}

function columnFromAgg(agg) {
  const aggTypeName = agg.type && agg.type.name;
  const aggParams = agg.params || {};
  const isAggregator = (agg.schema.group === 'buckets');

  const { field, customLabel } = aggParams;

  const fieldName = field ? field.displayName : '';
  const fieldLabel = `${fieldName}[${aggTypeName}]`;

  let type;

  switch (aggTypeName) {
    case 'count':
    case 'cardinality':
      type = 'integer';
      break;

    case 'range':
    case 'date_range':
      type = fieldType(field) + '_range';
      break;

    default:
      type = fieldType(field);
  }

  const defaultSort = _([
    'number', 'integer', 'date',
    'number_range', 'integer_range', 'date_range'
  ]).indexBy().mapValues(_.constant('asc')).value();

  // Show param will have 3 values: true, false, undefined.
  //
  // The 'undefined' value is used to identify non-user-explicit reasons to
  // hide the column, such as the column being not enabled or the agg being
  // partially defined.
  const show = agg.type && (agg.enabled || undefined);

  return {
    isAggregator,
    id: agg.id,
    label: customLabel || fieldLabel,
    type,
    scaleType: 'linear',
    sort: defaultSort[type] || 'none',
    show
  };
}

function getColumnValue(column, value) {
  switch (column.type) {
    case 'date':
    case 'date_range':
      // Dates are typically extracted from ES as numbers - cast them to Dates
      value = new Date(+value);
      break;

    default:
  }

  return value;
}

function getMetricValue(agg, column, bucket) {
  return getColumnValue(column, agg.getValue(bucket));
}

function getAggregatorValue(agg, column, bucket) {
  switch (agg.type.name) {
    case 'range':
    case 'date_range':
      // Note that ranges can be unbounded (no from, no to or *neither*)
      const hasFrom = bucket.hasOwnProperty('from');
      const hasTo = bucket.hasOwnProperty('to');

      let value;

      if (hasFrom && hasTo) { value = (bucket.from + bucket.to) / 2; }
      else if (hasFrom) { value = bucket.from; }
      else if (hasTo) { value = bucket.to; }
      else { value = 0; }

      const from = hasFrom ? getColumnValue(column, bucket.from) : null;
      const to = hasTo ? getColumnValue(column, bucket.to) : null;

      return { from, to, value };

    default:
      return getColumnValue(column, bucket.key);
  }
}

function addRecords(columns, aggs, records, aggregations) {
  const columnsById = _.indexBy(columns, 'id');
  const bucketAggs = _.filter(aggs.bySchemaGroup.buckets, 'enabled');

  function addRecordsRecurse(bucket, level, values) {
    if (level === bucketAggs.length) {
      if (!bucket.doc_count) { return; }

      const record = _(columns)
        .map(col => col.isAggregator
          ? values[col.id]
          : getMetricValue(aggs.byId[col.id], col, bucket))
        .concat([ bucket.doc_count || 0 ])
        .value();

      records.push(record);
      return;
    }

    const agg = bucketAggs[level++];
    const aggId = agg.id;
    const column = columnsById[aggId];
    const childBuckets = bucket[aggId].buckets;

    _.forEach(childBuckets, childBucket => {
      values[aggId] = getAggregatorValue(agg, column, childBucket);
      addRecordsRecurse(childBucket, level, values);
    });
  }

  return addRecordsRecurse(aggregations, 0, {});
}

function computeColumnRanges(columns, records) {
  return _.map(columns, function colToRange(col, c) {
    const values = _(records).map(c);

    switch (col.type) {
      case 'string':
        return (col.sort === 'none')
          ? values.uniq().reverse().value()     // ES sorts first rows as most relevant
          : values.sortByOrder(_.identity, col.sort).uniq(true).value();

      case 'ip':
        return (col.sort === 'none')
          ? values.uniq().reverse().value()
          : values.sortByOrder(paddedIp, col.sort).uniq(true).value();

      case 'number_range':
      case 'integer_range':
      case 'date_range':
        const sort = (col.sort === 'desc') ? 'desc' : 'asc';
        const samples = values.sortByOrder('value', sort).value();
        const lo = samples[0];
        const hi = samples[samples.length - 1];

        return {
          values: [ lo && lo.value, hi && hi.value ],
          samples
        };

      default:
        return (col.sort === 'desc')
          ? [ values.max(), values.min() ]
          : [ values.min(), values.max() ];
    }
  });
}


// Color Functions

function colorRangeFromRecords(columns, records) {
  const docsCountIdx = columns.length;
  const values = _(records).map(docsCountIdx);

  let min = values.min();
  let max = values.max();

  if (min === max) { min -= 1; max += 1; }

  return [ min, max ];
}


// Columns Visibility Functions

function visibleColumns(columns) {
  return _.filter(columns, 'show');
}

function visibleColRanges(columns, colRanges) {
  return _.filter(colRanges, (cr, c) => columns[c].show);
}

function visibleColumnsMap(columns) {
  let fc = 0;
  return _.map(columns, (col, c) => col.show ? fc++ : -1);
}

function allColumnsMap(columns) {
  return _(columns.length).range()
    .filter(c => columns[c].show)
    .value();
}

function visibleColumnsData(colsData) {
  const columns = colsData.columns;

  let result = {
    columns: visibleColumns(columns),
    colRanges: visibleColRanges(columns, colsData.colRanges),
    allToVisible: visibleColumnsMap(columns),
    visibleToAll: allColumnsMap(columns)
  };

  return result;
}


// Record Filtering Functions

function isLineFilteredIn(rd, colFilters, line) {
  const { columns, visibleToAll, scales } = rd;

  return _.every(columns, function (col, c) {
    let l = visibleToAll[c];

    let filter = colFilters[col.id];
    if (!filter) { return true; }

    let v = scales.columns[c](line[l]);
    return filter.v0 <= v && v <= filter.v1;
  });
}

function filteredData(rd, colFilters) {
  const { records } = rd;

  const filteredRecords = _.partition(records,
    record => isLineFilteredIn(rd, colFilters, record));

  return { records: filteredRecords };
}


export function columnsFromVis(vis, oldColumns = []) {
  let columns = _.map(vis.aggs, columnFromAgg);

  const oldDataById = _.reduce(oldColumns, function (memo, col, pos) {
    memo[col.id] = { col, pos };
    return memo;
  }, {});

  // Apply old columns data
  const copiedFields = ['scaleType', 'sort'];

  _.forEach(columns, col => {
    const oldData = oldDataById[col.id];
    if (!oldData) { return; }

    _.assign(col, _.pick(oldData.col, copiedFields));
    col.show = col.show && (oldData.col.show !== false);    // undefined => true
  });

  // Sort columns by position
  columns = _.sortBy(columns, col => {
    const oldData = oldDataById[col.id];
    return oldData ? oldData.pos : Number.MAX_SAFE_INTEGER;
  });

  return columns;
}

export function recordsFromResponse(columns, aggs, resp) {
  if (!aggs || !resp) { return []; }

  const { aggregations } = resp;
  if (!aggregations) { return []; }

  const records = [];
  addRecords(columns, aggs, records, aggregations);

  return records;
}

export default class RenderData {
  constructor(data = {}) {
    data = _.defaults(data, {
      params: {},
      columns: [],
      records: []
    });

    const { columns, records, colors } = data;

    const colRanges = computeColumnRanges(columns, records);
    const colorRange = colorRangeFromRecords(columns, records);

    _.assign(this, data, {
      colorRange,

      all: {
        columns, colRanges
      },

      filtered: { records: [records, []] },

      view: {
        // View components are the only ones mutable by charts
        hover: {},
        colFilters: {}
      }
    });

    this.applyVisibility();
  }

  applyColumnFilters() {
    _.assign(this.filtered, filteredData(this, this.view.colFilters));
  }

  applyVisibility() {
    _.assign(this, visibleColumnsData(this.all));
    this.scales = scalesFromRenderData(this);
  }

  moveColumn(srcIdx, dstIdx) {
    srcIdx = this.visibleToAll[srcIdx];
    dstIdx = this.visibleToAll[dstIdx];

    const { columns, colRanges } = this.all;

    arrayMoveValue(columns, srcIdx, dstIdx);
    arrayMoveValue(colRanges, srcIdx, dstIdx);

    _.forEach(this.records, rec => arrayMoveValue(rec, srcIdx, dstIdx));

    this.applyVisibility();
  }

  sortColumn(colIdx) {
    const column = this.columns[colIdx];
    let next;

    switch (column.type) {
      case 'string':
      case 'ip':
        next = { asc: 'desc', desc: 'none', none: 'asc' };
        break;

      default:
        next = { asc: 'desc', desc: 'asc' };
        break;
    }

    column.sort = next[column.sort] || 'asc';

    this.all.colRanges = computeColumnRanges(this.columns, this.records);
    delete this.view.colFilters[column.id];

    this.applyVisibility();
    this.applyColumnFilters();
  }

  setColumnVisible(colIdx, visible) {
    const column = this.columns[colIdx];
    column.show = visible;

    this.applyVisibility();
  }

  toggleColumnVisible(colIdx) {
    const column = this.columns[colIdx];
    column.show = !column.show;

    this.applyVisibility();
  }
};

