
import * as cc from './chartcommons';
import { assert } from './utils';

import * as d3 from './d3';
import _ from 'lodash';

import '../styles/scatterplot.css';


// Consts

const fontSize = 14;

const pointsHoverRadius = 6;

const filterRangeSize = 16;
const filterRangeSizeHalf = filterRangeSize / 2;
const filterRangeHandlesSize = 5;
const filterValueRadius = 5;
const filterTextPad = 4;
const filterTextCrossPad = filterRangeSizeHalf + 4;

const padLeft = 60;
const padRight = 60;
const padTop = 20;
const padBottom = 60;

const padChartX0 = 12;
const padChartY0 = 10;
const padChartX1 = padChartX0;
const padChartY1 = 6;
const axisTickSize = 4;
const axisLabelPad = 10;
const tooltipPad = 30;


// Local Functions

// Geometry

function valueToX(ctx, val) {
  return ctx.x0Chart + val * ctx.contentBox[2];
}

function xToValue(ctx, x) {
  return (x - ctx.x0Chart) / ctx.contentBox[2];
}

function valueToY(ctx, val) {
  return ctx.y0Chart - val * ctx.contentBox[3];
}

function yToValue(ctx, y) {
  return (ctx.y0Chart - y) / ctx.contentBox[3];
}

const toValue = { x: xToValue, y: yToValue };

function cx(ctx) {
  const c = ctx.scatterplotAxes.x;
  const l = ctx.visibleToAll[c];

  return line => ctx.xScale(line[l]);
}

function cy(ctx) {
  const c = ctx.scatterplotAxes.y;
  const l = ctx.visibleToAll[c];

  return line => ctx.yScale(line[l]);
}

function color(ctx) {
  return line => ctx.scales.color(line[line.length - 1]);
}

function text(ctx, line) {
  return c => {
    const colType = ctx.columns[c].type;
    const colRange = ctx.colRanges[c];

    return cc.sampleToText(colType, colRange, line[ctx.visibleToAll[c]]);
  };
}

function closestLine(ctx, posX, posY) {
  const px = cx(ctx);
  const py = cy(ctx);

  return cc.closestPoint(ctx, px, py, posX, posY);
}

function chartPosToValues(ctx, pos) {
  return [
    (pos[0] - ctx.x0Chart) / (ctx.x1Chart - ctx.x0Chart),
    (pos[1] - ctx.y0Chart) / (ctx.y1Chart - ctx.y0Chart)
  ];
}


// Interaction

function updateHover(ctx) {
  const mPos = cc.rawMousePos(ctx, d3.event);
  if (!cc.rectContains(ctx.hoverRect, mPos)) { return cc.resetHover(ctx); }

  const hover = closestLine(ctx, mPos[0], mPos[1]);
  if (hover.line === ctx.view.hover.line) { return; }

  hover.event = d3.event;

  ctx.view.hover = hover;
  ctx.viewChanges.hover = true;

  cc.refresh(ctx);
}


function filterDefineFilter(ctx) {
  return !d3.event.button && cc.rectContains(ctx.filterDefineRect, d3.mouse(ctx.svg));
}

function filterDefineStart(ctx) {
  const mPos = [d3.event.x, d3.event.y];
  const values = chartPosToValues(ctx, mPos);

  const xc = ctx.scatterplotAxes.x;
  const yc = ctx.scatterplotAxes.y;

  const xColName = ctx.columns[xc].id;
  const yColName = ctx.columns[yc].id;

  const xFilter = ctx.view.colFilters[xColName];
  const yFilter = ctx.view.colFilters[yColName];

  ctx.mousePress = {
    x: { v: values[0], p: mPos[0], filter: xFilter },
    y: { v: values[1], p: mPos[1], filter: yFilter }
  };

  cc.resetHover(ctx);
}

function filterDefineAxisMove(ctx, axis) {
  const p = d3.event[axis];
  const v = toValue[axis](ctx, p);
  const vPress = ctx.mousePress[axis].v;

  const c = ctx.scatterplotAxes[axis];
  const colName = ctx.columns[c].id;

  if (vPress >= 0) {
    const v0 = Math.min(v, vPress);
    const v1 = Math.max(v, vPress);

    ctx.view.colFilters[colName] = cc.rangeFilter(ctx, c, v0, v1);
  } else {
    const filter = ctx.mousePress[axis].filter;
    if (filter) { ctx.view.colFilters[colName] = filter; }
    else { delete ctx.view.colFilters[colName]; }
  }

  ctx.viewChanges.colFilters = true;
}

function filterDefineMove(ctx) {
  filterDefineAxisMove(ctx, 'x');
  filterDefineAxisMove(ctx, 'y');

  cc.refresh(ctx);
}

function rangeFilterDragStart(ctx, axis) {
  const colName = ctx.columns[ctx.scatterplotAxes[axis]].id;
  const filter = ctx.view.colFilters[colName];

  ctx.mousePress = {
    v0: filter.v0,
    v1: filter.v1,
    vPress: toValue[axis](ctx, d3.event[axis])
  };
}

function rangeFilterDragMove(ctx, axis) {
  const { v0, v1, vPress } = ctx.mousePress;
  const drag = toValue[axis](ctx, d3.event[axis]) - vPress;

  const c = ctx.scatterplotAxes[axis];
  const colName = ctx.columns[c].id;

  ctx.view.colFilters[colName] = cc.rangeFilter(ctx, c, v0 + drag, v1 + drag);

  ctx.viewChanges.colFilters = true;
  cc.refresh(ctx);
}

function rangeFilterMinDragMove(ctx, axis) {
  const { v0, v1, vPress } = ctx.mousePress;
  const drag = toValue[axis](ctx, d3.event[axis]) - vPress;

  const c = ctx.scatterplotAxes[axis];
  const colName = ctx.columns[c].id;

  ctx.view.colFilters[colName] = cc.rangeFilter(
    ctx, c, Math.min(v0 + drag, v1), v1);

  ctx.viewChanges.colFilters = true;
  cc.refresh(ctx);
}

function rangeFilterMaxDragMove(ctx, axis) {
  const { v0, v1, vPress } = ctx.mousePress;
  const drag = toValue[axis](ctx, d3.event[axis]) - vPress;

  const c = ctx.scatterplotAxes[axis];
  const colName = ctx.columns[c].id;

  ctx.view.colFilters[colName] = cc.rangeFilter(
    ctx, c, v0, Math.max(v0, v1 + drag));

  ctx.viewChanges.colFilters = true;
  cc.refresh(ctx);
}


function initMouse(ctx) {
  ctx.filterDefineEnd = () => cc.filterDefineEnd(ctx);

  ctx.xRangeFilterDragStart = () => rangeFilterDragStart(ctx, 'x');
  ctx.xRangeFilterDragMove = () => rangeFilterDragMove(ctx, 'x');
  ctx.xRangeFilterMinDragMove = () => rangeFilterMinDragMove(ctx, 'x');
  ctx.xRangeFilterMaxDragMove = () => rangeFilterMaxDragMove(ctx, 'x');

  ctx.yRangeFilterDragStart = () => rangeFilterDragStart(ctx, 'y');
  ctx.yRangeFilterDragMove = () => rangeFilterDragMove(ctx, 'y');
  ctx.yRangeFilterMinDragMove = () => rangeFilterMinDragMove(ctx, 'y');
  ctx.yRangeFilterMaxDragMove = () => rangeFilterMaxDragMove(ctx, 'y');

  d3.select(ctx.svg)
    .on('mousemove', () => updateHover(ctx))
    .call(d3.drag().container(ctx.svg)
      .filter(() => filterDefineFilter(ctx))
      .on('start', () => filterDefineStart(ctx))
      .on('drag', () => filterDefineMove(ctx))
      .on('end', ctx.filterDefineEnd));
}


// Rendering

function renderPoints(ctx, group) {
  const { pointsRadius } = ctx;

  const filteredLines = ctx.filtered.records;
  const filteredOutLines = ctx.showFilteredOutRecords ? filteredLines[1] : [];

  const x = cx(ctx);
  const y = cy(ctx);


  const outGroup = cc.materialize(group, 'g', 'prl-bsp-filtered-out', [0]);

  cc.materialize(outGroup, 'circle', '', filteredOutLines)
    .attr('cx', x)
    .attr('cy', y)
    .attr('r', pointsRadius);


  const inGroup = cc.materialize(group, 'g', 'prl-bsp-filtered-in', [0]);

  cc.materialize(inGroup, 'circle', '', filteredLines[0])
    .attr('cx', x)
    .attr('cy', y)
    .attr('r', pointsRadius)
    .attr('fill', line => ctx.scales.color(line[line.length - 1]));
}

function renderContentGroup(ctx, svg) {
  const contentGroup = cc.materialize(svg, 'g', 'prl-bsp-content-group', [0]);

  renderPoints(ctx, contentGroup);
}


function renderAxis(ctx, group) {
  const { xAxis, yAxis, xScale, yScale } = ctx;

  const xc = ctx.scatterplotAxes.x;
  const yc = ctx.scatterplotAxes.y;
  const xColumn = ctx.columns[xc];
  const yColumn = ctx.columns[yc];
  const xColRange = ctx.colRanges[xc];
  const yColRange = ctx.colRanges[yc];


  const axisGroup = cc.materialize(group, 'g', 'prl-bsp-axis-group', [0]);

  // X Axis

  const xAxisGroup = cc.materialize(axisGroup, 'g', 'prl-bsp-xaxis-group', [0]);
  const xTicks = cc.columnTicks(
    xColumn, ctx.scales.columns[xc], ctx.refSizeX, ctx.ticksDistance);

  cc.materialize(xAxisGroup, 'line', 'prl-bsp-axis', [0])
    .attr('x1', xAxis).attr('x2', ctx.x1Chart)
    .attr('y1', yAxis).attr('y2', yAxis);

  cc.materialize(xAxisGroup, 'line', 'prl-bsp-tick-line', xTicks)
    .attr('x1', xScale).attr('x2', xScale)
    .attr('y1', yAxis).attr('y2', yAxis + axisTickSize);

  cc.materialize(xAxisGroup, 'text', 'prl-bsp-tick-text', xTicks)
    .attr('x', xScale)
    .attr('y', yAxis + axisTickSize + 2)
    .attr('dy', '0.25em')
    .attr('transform', t => `rotate(90 ${xScale(t)},${yAxis + axisTickSize + 2})`)
    .text(s => cc.sampleToText(xColumn.type, xColRange, s));

  // Y Axis

  const yAxisGroup = cc.materialize(axisGroup, 'g', 'prl-bsp-yaxis-group', [0]);
  const yTicks = cc.columnTicks(
    yColumn, ctx.scales.columns[yc], ctx.refSizeY, ctx.ticksDistance);

  cc.materialize(yAxisGroup, 'line', 'prl-bsp-axis', [0])
    .attr('x1', xAxis).attr('x2', xAxis)
    .attr('y1', yAxis).attr('y2', ctx.y1Chart);

  cc.materialize(yAxisGroup, 'line', 'prl-bsp-tick-line', yTicks)
    .attr('x1', xAxis).attr('x2', xAxis - axisTickSize)
    .attr('y1', yScale).attr('y2', yScale);

  cc.materialize(yAxisGroup, 'text', 'prl-bsp-tick-text', yTicks)
    .attr('x', xAxis - axisTickSize - 2)
    .attr('y', yScale)
    .attr('dy', '0.25em')
    .text(s => cc.sampleToText(yColumn.type, yColRange, s));
}

function renderLabels(ctx, group) {
  const labelsGroup = cc.materialize(group, 'g', 'prl-bsp-labels-group', [0]);

  let x = .5 * (ctx.x0Chart + ctx.x1Chart);
  let y = ctx.y1Chart - axisLabelPad;

  cc.materialize(labelsGroup, 'text', 'prl-bsp-axis-x-label', [0])
    .attr('x', x).attr('y', y)
    .attr('font-size', fontSize)
    .text(ctx.columns[ctx.scatterplotAxes.x].label);

  x = ctx.x1Chart + axisLabelPad;
  y = .5 * (ctx.y0Chart + ctx.y1Chart);

  cc.materialize(labelsGroup, 'text', 'prl-bsp-axis-y-label', [0])
    .attr('x', x).attr('y', y)
    .attr('font-size', fontSize)
    .attr('transform', `rotate(90 ${x},${y})`)
    .text(ctx.columns[ctx.scatterplotAxes.y].label);
}

function tooltipHtml(ctx, line) {
  const xc = ctx.scatterplotAxes.x;
  const yc = ctx.scatterplotAxes.y;

  const txt = text(ctx, line);

  let cs = [xc, yc];
  cs = cs.concat(_(0).range(ctx.columns.length)
    .difference(cs)
    .value());

  const tRows = _(cs)
    .map(c => `<tr><td>${ctx.columns[c].label}</td><td>${txt(c)}</td></tr>`)
    .tap(arr => arr.splice(2, 0,
      `<tr><td>Count</td><td>${line[line.length - 1]}</td></tr>`))
    .join('');

  return `<table><tbody>${tRows}</tbody></table>`;
}

function renderHover(ctx, group) {
  const { pointsHoverRadius } = ctx;

  const hoverLine = ctx.view.hover.line ? [ctx.view.hover.line] : [];

  const x = cx(ctx);
  const y = cy(ctx);

  const hoverColor = ctx.colors.highlight || 'black';


  const hoverGroup = cc.materialize(group, 'g', 'prl-bsp-hover-group', [0]);

  // Projections to axis

  cc.materialize(hoverGroup, 'line', 'prl-bsp-hover-proj-h', hoverLine)
    .attr('x1', ctx.xAxis).attr('x2', x)
    .attr('y1', y).attr('y2', y)
    .attr('stroke', hoverColor);
  cc.materialize(hoverGroup, 'line', 'prl-bsp-hover-proj-v', hoverLine)
    .attr('x1', x).attr('x2', x)
    .attr('y1', ctx.yAxis).attr('y2', y)
    .attr('stroke', hoverColor);

  // Hovered point

  cc.materialize(hoverGroup, 'circle', 'prl-bsp-hover', hoverLine)
    .attr('cx', x).attr('cy', y)
    .attr('r', pointsHoverRadius)
    .attr('stroke', hoverColor);
}

function renderTooltip(ctx, group) {
  const hoverLine = ctx.view.hover.line ? [ctx.view.hover.line] : [];

  const x = cx(ctx);
  const y = cy(ctx);

  const body = d3.select('body');

  const svgRect = ctx.svg.getBoundingClientRect();
  const svgL = window.pageXOffset + svgRect.left;
  const svgT = window.pageYOffset + svgRect.top;
  const svgR = body.node().clientWidth - window.pageXOffset - svgRect.right;
  const svgB = body.node().clientHeight - window.pageYOffset - svgRect.bottom;

  const chartPageL = svgL - ctx.viewRect[0];
  const chartPageT = svgT - ctx.viewRect[1];
  const chartPageR = svgR + svgRect.width + ctx.viewRect[0];
  const chartPageB = svgB + svgRect.height + ctx.viewRect[1];


  cc.materialize(body, 'div', 'prl-bsp-tooltip', hoverLine)
    .each(function (line) {
      const toolTip = d3.select(this);

      const xPos = x(line);
      const yPos = y(line);

      const xRelPos = (xPos - ctx.x0Chart) / ctx.contentBox[2];
      const yRelPos = (yPos - ctx.y1Chart) / ctx.contentBox[3];

      const positions = {
        left: undefined, right: undefined,
        top: undefined, bottom: undefined
      };

      if (xRelPos < .5) {
        positions.left = `${chartPageL + xPos + tooltipPad}px`;
      } else {
        positions.right = `${chartPageR - xPos + tooltipPad}px`;
      }

      if (yRelPos < .5) {
        positions.top = `${chartPageT + yPos + tooltipPad}px`;
      } else {
        positions.bottom = `${chartPageB - yPos + tooltipPad}px`;
      }

      toolTip
        .style('left', positions.left)
        .style('top', positions.top)
        .style('right', positions.right)
        .style('bottom', positions.bottom)
        .html(tooltipHtml(ctx, line));
    });
}

function renderFilters(ctx, group) {
  const xc = ctx.scatterplotAxes.x;
  const yc = ctx.scatterplotAxes.y;

  const xCol = ctx.columns[xc];
  const yCol = ctx.columns[yc];

  const xFilter = ctx.view.colFilters[xCol.id];
  const yFilter = ctx.view.colFilters[yCol.id];

  const filters = (flt, type) => flt && flt.type === type ? [flt] : [];

  const xRangeFilters = filters(xFilter, 'range');
  const yRangeFilters = filters(yFilter, 'range');
  const xValueFilters = filters(xFilter, 'value');
  const yValueFilters = filters(yFilter, 'value');


  cc.materialize(group, 'g', 'prl-bsp-xfilter-range-group', xRangeFilters)
    .each(ctx.renderXRangeFilter);

  cc.materialize(group, 'g', 'prl-bsp-yfilter-range-group', yRangeFilters)
    .each(ctx.renderYRangeFilter);

  cc.materialize(group, 'g', 'prl-bsp-xfilter-value-group', xValueFilters)
    .each(ctx.renderXValueFilter);

  cc.materialize(group, 'g', 'prl-bsp-yfilter-value-group', yValueFilters)
    .each(ctx.renderYValueFilter);
}

function renderXRangeFilter(ctx, filter) {
  const group = d3.select(this);

  const c = ctx.scatterplotAxes.x;
  const colType = ctx.columns[c].type;
  const colRange = ctx.colRanges[c];

  const x0 = valueToX(ctx, filter.v0);
  const x1 = valueToX(ctx, filter.v1);
  const y0 = ctx.yAxis + filterRangeSizeHalf;
  const y1 = y0 - filterRangeSize;
  const width = x1 - x0;
  const handlesSize = Math.min(width, filterRangeHandlesSize);
  const x0Text = x0 - filterTextPad;
  const x1Text = x1 + filterTextPad;
  const yText = ctx.yAxis + filterTextCrossPad;


  cc.materialize(group, 'rect', 'prl-bsp-filter-range', [0])
    .attr('x', x0).attr('width', width)
    .attr('y', y1).attr('height', filterRangeSize)
    .call(d3.drag().container(ctx.svg)
      .on('start', ctx.xRangeFilterDragStart)
      .on('drag', ctx.xRangeFilterDragMove)
      .on('end', ctx.filterDefineEnd));

  cc.materialize(group, 'rect', 'prl-bsp-filter-range-min', [0])
    .attr('x', x0).attr('width', handlesSize)
    .attr('y', y1).attr('height', filterRangeSize)
    .call(d3.drag().container(ctx.svg)
      .on('start', ctx.xRangeFilterDragStart)
      .on('drag', ctx.xRangeFilterMinDragMove)
      .on('end', ctx.filterDefineEnd));

  cc.materialize(group, 'rect', 'prl-bsp-filter-range-max', [0])
    .attr('x', Math.max(x1 - filterRangeHandlesSize, x0))
    .attr('width', handlesSize)
    .attr('y', y1)
    .attr('height', filterRangeSize)
    .call(d3.drag().container(ctx.svg)
      .on('start', ctx.xRangeFilterDragStart)
      .on('drag', ctx.xRangeFilterMaxDragMove)
      .on('end', ctx.filterDefineEnd));


  cc.materialize(group, 'line', 'prl-bsp-filter-line', [x0, x1])
    .attr('x1', _.identity).attr('x2', _.identity)
    .attr('y1', y1).attr('y2', ctx.y1Chart);

  cc.materialize(group, 'text', 'prl-bsp-filter-text-min', [0])
    .attr('x', x0Text).attr('y', yText)
    .attr('transform', `rotate(90 ${x0Text},${yText})`)
    .attr('dy', '1em')
    .text(cc.sampleToText(colType, colRange, filter.s0));

  cc.materialize(group, 'text', 'prl-bsp-filter-text-max', [0])
    .attr('x', x1Text).attr('y', yText)
    .attr('transform', `rotate(90 ${x1Text},${yText})`)
    .text(cc.sampleToText(colType, colRange, filter.s1));
}

function renderYRangeFilter(ctx, filter) {
  const group = d3.select(this);

  const c = ctx.scatterplotAxes.y;
  const colType = ctx.columns[c].type;
  const colRange = ctx.colRanges[c];

  const x0 = ctx.xAxis - filterRangeSizeHalf;
  const x1 = x0 + filterRangeSize;
  const y0 = valueToY(ctx, filter.v0);
  const y1 = valueToY(ctx, filter.v1);
  const height = y0 - y1;
  const handlesSize = Math.min(height, filterRangeHandlesSize);
  const y0Text = y0 + filterTextPad;
  const y1Text = y1 - filterTextPad;
  const xText = ctx.xAxis - filterTextCrossPad;


  cc.materialize(group, 'rect', 'prl-bsp-filter-range', [0])
    .attr('x', x0).attr('width', filterRangeSize)
    .attr('y', y1).attr('height', height)
    .call(d3.drag().container(ctx.svg)
      .on('start', ctx.yRangeFilterDragStart)
      .on('drag', ctx.yRangeFilterDragMove)
      .on('end', ctx.filterDefineEnd));

  cc.materialize(group, 'rect', 'prl-bsp-filter-range-min', [0])
    .attr('x', x0)
    .attr('width', filterRangeSize)
    .attr('y', Math.max(y0 - filterRangeHandlesSize, y1))
    .attr('height', handlesSize)
    .call(d3.drag().container(ctx.svg)
      .on('start', ctx.yRangeFilterDragStart)
      .on('drag', ctx.yRangeFilterMinDragMove)
      .on('end', ctx.filterDefineEnd));

  cc.materialize(group, 'rect', 'prl-bsp-filter-range-max', [0])
    .attr('x', x0).attr('width', filterRangeSize)
    .attr('y', y1).attr('height', handlesSize)
    .call(d3.drag().container(ctx.svg)
      .on('start', ctx.yRangeFilterDragStart)
      .on('drag', ctx.yRangeFilterMaxDragMove)
      .on('end', ctx.filterDefineEnd));


  cc.materialize(group, 'line', 'prl-bsp-filter-line', [y0, y1])
    .attr('x1', x1).attr('x2', ctx.x1Chart)
    .attr('y1', _.identity).attr('y2', _.identity);

  cc.materialize(group, 'text', 'prl-bsp-filter-text-min', [0])
    .attr('x', xText).attr('y', y0Text)
    .attr('dy', '1em')
    .text(cc.sampleToText(colType, colRange, filter.s0));

  cc.materialize(group, 'text', 'prl-bsp-filter-text-max', [0])
    .attr('x', xText).attr('y', y1Text)
    .text(cc.sampleToText(colType, colRange, filter.s1));
}

function renderXValueFilter(ctx, filter) {
  const group = d3.select(this);

  const c = ctx.scatterplotAxes.x;
  const colType = ctx.columns[c].type;
  const colRange = ctx.colRanges[c];

  const x = valueToX(ctx, filter.v0);
  const y0 = ctx.yAxis + filterRangeSizeHalf;
  const yText = ctx.yAxis + filterTextCrossPad;


  cc.materialize(group, 'circle', 'prl-bsp-filter-value', [0])
    .attr('cx', x).attr('cy', ctx.yAxis)
    .attr('r', filterValueRadius);


  cc.materialize(group, 'line', 'prl-bsp-filter-line', [0])
    .attr('x1', x).attr('x2', x)
    .attr('y1', y0).attr('y2', ctx.y1Chart);

  cc.materialize(group, 'text', 'prl-bsp-filter-text', [0])
    .attr('x', x).attr('y', yText)
    .attr('transform', `rotate(90 ${x},${yText})`)
    .attr('dy', '.5em')
    .text(cc.sampleToText(colType, colRange, filter.s1));
}

function renderYValueFilter(ctx, filter) {
  const group = d3.select(this);

  const c = ctx.scatterplotAxes.y;
  const colType = ctx.columns[c].type;
  const colRange = ctx.colRanges[c];

  const x0 = ctx.xAxis + filterRangeSizeHalf;
  const y = valueToY(ctx, filter.v0);
  const xText = ctx.xAxis - filterTextCrossPad;


  cc.materialize(group, 'circle', 'prl-bsp-filter-value', [0])
    .attr('cx', ctx.xAxis).attr('cy', y)
    .attr('r', filterValueRadius);


  cc.materialize(group, 'line', 'prl-bsp-filter-line', [0])
    .attr('x1', x0).attr('x2', ctx.x1Chart)
    .attr('y1', y).attr('y2', y);

  cc.materialize(group, 'text', 'prl-bsp-filter-text', [0])
    .attr('x', xText).attr('y', y)
    .attr('dy', '.5em')
    .text(cc.sampleToText(colType, colRange, filter.s1));
}

function renderCloseButton(ctx, group) {
  cc.materialize(group, 'text', 'prl-bsp-btn-close', [0])
    .attr('x', ctx.refSizeX + padRight - 10).attr('y', 10)
    .text('\uf066')
    .on('click', ctx.onScatterplotClose);
}

function renderOverlayGroup(ctx, svg) {
  const overlayGroup = cc.materialize(svg, 'g', 'prl-bsp-overlay-group', [0]);

  renderAxis(ctx, overlayGroup);
  renderLabels(ctx, overlayGroup);
  renderHover(ctx, overlayGroup);
  renderTooltip(ctx, overlayGroup);
  renderFilters(ctx, overlayGroup);
  renderCloseButton(ctx, overlayGroup);
}

function renderScatterPlot(ctx, renderContent = true, renderOverlay = true) {
  const svg = d3.select(ctx.svg);

  if (renderContent) { renderContentGroup(ctx, svg); }
  if (renderOverlay) { renderOverlayGroup(ctx, svg); }
}


// State

function buildSizes(ctx) {
  const result = {};

  const containerRect = ctx.container.getBoundingClientRect();

  let { pointsRadius } = ctx;
  pointsRadius = pointsRadius || 3;

  const axisPad = filterRangeSizeHalf + pointsRadius + 2;

  result.pointsRadius = pointsRadius;
  result.pointsHoverRadius = pointsRadius + 3;

  result.sizeX = containerRect.width;
  result.sizeY = containerRect.height;

  result.refSizeX = Math.max(result.sizeX - padLeft - padRight, 1);
  result.refSizeY = Math.max(result.sizeY - padTop - padBottom, 1);

  result.xAxis = padChartX0 + .5;
  result.yAxis = Math.round(result.refSizeY - padChartY0 - fontSize) + .5;

  result.x0Chart = result.xAxis + axisPad;
  result.x1Chart = Math.round(result.refSizeX - padChartX1) + .5;

  result.y0Chart = result.yAxis - axisPad;
  result.y1Chart = padChartY1 + .5;

  result.xMidChart = .5 * (result.x0Chart + result.x1Chart);
  result.yMidChart = .5 * (result.y0Chart + result.y1Chart);

  result.viewBox = [ -padLeft, -padTop, result.sizeX, result.sizeY ];
  result.viewRect = [ -padLeft, -padTop, result.sizeX - padLeft, result.sizeY - padTop ];

  result.contentBox = [
    result.x0Chart, result.y1Chart,
    result.x1Chart - result.x0Chart, result.y0Chart - result.y1Chart
  ];

  result.hoverRect = [
    result.x0Chart - pointsRadius, result.y1Chart - pointsRadius,
    result.x1Chart + pointsRadius, result.y0Chart + pointsRadius
  ];

  result.filterDefineRect = [
    result.xAxis - filterRangeSizeHalf, result.y1Chart,
    result.x1Chart, result.yAxis + filterRangeSizeHalf
  ];

  return result;
}

function buildScales(ctx) {
  const x0 = ctx.x0Chart;
  const y0 = ctx.y0Chart;
  const xLen = ctx.contentBox[2];
  const yLen = ctx.contentBox[3];

  const axes = ctx.scatterplotAxes;
  assert(axes);

  const vxScale = ctx.scales.columns[axes.x];
  const vyScale = ctx.scales.columns[axes.y];

  return {
    xScale: val => x0 + xLen * vxScale(val),
    yScale: val => y0 - yLen * vyScale(val)
  };
}

function initState(base, ctx) {
  ctx = _.assign({
    viewChanges: {},

    mousePress: null,
    filterRect: null
  }, base, ctx);

  _.assign(ctx, {
    renderXRangeFilter: _.partial(renderXRangeFilter, ctx),
    renderYRangeFilter: _.partial(renderYRangeFilter, ctx),
    renderXValueFilter: _.partial(renderXValueFilter, ctx),
    renderYValueFilter: _.partial(renderYValueFilter, ctx),
  });

  _.assign(ctx, buildSizes(ctx));
  _.assign(ctx, buildScales(ctx));

  return ctx;
}


// Exports

export default class ScatterPlot {
  constructor(opts) {
    this.baseCtx = opts;
    this.ctx = null;
  }

  render(ctx) {
    this.ctx = ctx = initState(this.baseCtx, ctx);

    d3.select(ctx.svg)
      .attr('width', ctx.sizeX)
      .attr('height', ctx.sizeY)
      .attr('viewBox', ctx.viewBox.join(' '));

    renderScatterPlot(ctx);

    initMouse(ctx);
  }

  refresh(viewChanges) {
    if (!this.ctx) { return; }

    let refreshContent = false;
    let refreshOverlay = false;

    _.chain(viewChanges).keys().forEach(viewData => {
      switch (viewData) {
        case 'colFilters': refreshContent = true;
        case 'hover': refreshOverlay = true;
      }
    }).commit();

    renderScatterPlot(this.ctx, refreshContent, refreshOverlay);
  }

  reset() {
    if (!this.ctx) { return; }

    d3.select(this.ctx.svg).selectAll('*').remove();
    d3.select('.prl-bsp-tooltip').remove();

    this.ctx = null;
  }
};

