
import * as cc from './chartcommons';

import * as d3 from './d3';
import _ from 'lodash';

import '../styles/scatterstrip.css';


const padChartX0 = 12;
const padChartY0 = 10;
const padChartX1 = padChartX0;
const padChartY1 = 6;
const axisLabelPadX = 5;
const axisLabelPadY = 3;
const axisPad = 5;
const fontSize = 14;

const pointsRadius = 1.5;
const pointsHoverRadius = 3;
const pointsHaloRadius = 5;


// Local Functions

// Geometry

function xToColumnIndex(ctx, x) {
  return Math.floor((ctx.columns.length - 1) * (x / ctx.refSizeX));
}

function mouseToChartPos(ctx, c, mPos) {
  let x0 = c * ctx.xStep;

  return cc.movedInsideRect(ctx.chartRect, [ mPos[0] - x0, mPos[1] ]);
}

function chartPosToValues(ctx, pos) {
  return [
    (pos[0] - ctx.x0Chart) / (ctx.x1Chart - ctx.x0Chart),
    (pos[1] - ctx.y0Chart) / (ctx.y1Chart - ctx.y0Chart)
  ];
}

function cx(ctx, c) {
  const l = ctx.visibleToAll[c + 1];
  return line => ctx.xScales[c](line[l]);
}

function cy(ctx, c) {
  const l = ctx.visibleToAll[c];
  return line => ctx.yScales[c](line[l]);
}

function closestLine(ctx, posX, posY) {
  const c = Math.min(xToColumnIndex(ctx, posX), ctx.columns.length - 2);
  if (c < 0) { return {}; }

  const x0 = c * ctx.xStep;
  const px = cx(ctx, c);
  const py = cy(ctx, c);

  return cc.closestPoint(ctx, px, py, posX - x0, posY);
}


// Interaction

function updateHover(ctx) {
  const mPos = cc.rawMousePos(ctx, d3.event);
  if (!cc.rectContains(ctx.contentRect, mPos)) { return cc.resetHover(ctx); }

  const hover = closestLine(ctx, mPos[0], mPos[1]);
  if (hover.line === ctx.view.hover.line) { return; }

  hover.event = d3.event;

  ctx.view.hover = hover;
  ctx.viewChanges.hover = true;

  cc.refresh(ctx);
}


function filterDefineFilter(ctx) {
  return !d3.event.button && cc.rectContains(ctx.contentRect, d3.mouse(ctx.svg));
}

function filterDefineStart(ctx) {
  const mPos = [d3.event.x, d3.event.y];
  const c = xToColumnIndex(ctx, mPos[0]);

  const chartPos = mouseToChartPos(ctx, c, mPos);
  const values = chartPosToValues(ctx, chartPos);

  ctx.mousePress = {
    c,
    vx: values[0], vy: values[1],
    x: chartPos[0], y: chartPos[1]
  };

  cc.resetHover(ctx);
}

function filterDefineMove(ctx) {
  const mPos = [d3.event.x, d3.event.y];

  const cy = ctx.mousePress.c;
  const cx = cy + 1;

  const chartPos = mouseToChartPos(ctx, cy, mPos);
  const values = chartPosToValues(ctx, chartPos);

  const cxName = ctx.columns[cx].id;
  const cyName = ctx.columns[cy].id;

  const v0x = Math.min(values[0], ctx.mousePress.vx);
  const v1x = Math.max(values[0], ctx.mousePress.vx);

  const v0y = Math.min(values[1], ctx.mousePress.vy);
  const v1y = Math.max(values[1], ctx.mousePress.vy);

  ctx.view.colFilters[cxName] = cc.rangeFilter(ctx, cx, v0x, v1x);
  ctx.view.colFilters[cyName] = cc.rangeFilter(ctx, cy, v0y, v1y);

  ctx.filterRect = {
    x0: _.min([ctx.mousePress.x, chartPos[0]]),
    x1: _.max([ctx.mousePress.x, chartPos[0]]),
    y0: _.min([ctx.mousePress.y, chartPos[1]]),
    y1: _.max([ctx.mousePress.y, chartPos[1]])
  };

  ctx.viewChanges.colFilters = true;
  cc.refresh(ctx);
}

function filterDefineEnd(ctx) {
  ctx.filterRect = null;
  cc.filterDefineEnd(ctx);
}

function scatterplotExpand(ctx, c) {
  d3.event.stopPropagation();
  ctx.onScatterplotExpand({ x: c + 1, y: c });
}


function initMouse(ctx) {
  ctx.filterDefineEnd = () => filterDefineEnd(ctx);
  ctx.scatterplotExpand = c => scatterplotExpand(ctx, c);

  d3.select(ctx.svg)
    .on('mousemove', () => updateHover(ctx))
    .on('mouseleave', () => cc.resetHover(ctx))
    .call(d3.drag()
      .container(ctx.svg)
      .filter(() => filterDefineFilter(ctx))
      .on('start', () => filterDefineStart(ctx))
      .on('drag', () => filterDefineMove(ctx))
      .on('end', ctx.filterDefineEnd));
}


// Rendering

function renderChartsStrip(ctx, renderContent = true, renderOverlay = true) {
  const svg = d3.select(ctx.svg);
  const canvas = ctx.canvas.getContext('2d');
  const columns = ctx.columns;
  const colIndexes = _.range(0, columns.length - 1, 1);

  cc.materialize(svg, 'g', 'ssg-chart', colIndexes)
    .attr('transform', cc.translator(ctx.refSizeX, columns.length))
    .each(function (c) {
      let chart = d3.select(this);

      if (renderContent) { renderContentGroupSvg(ctx, chart, c); }
      if (renderOverlay) { renderOverlayGroup(ctx, chart, c); }
    });

  if (ctx.useCanvas && renderContent) {
    canvas.setTransform(1, 0, 0, 1, 0, 0);
    canvas.clearRect(0, 0, ctx.sizeX, ctx.sizeY);

    _.forEach(colIndexes, function (c) {
      renderContentGroupCanvas(ctx, canvas, c);
    });
  }
}

function renderContentGroupSvg(ctx, group, c) {
  let contentGroup = cc.materialize(group, 'g', 'ssg-content', [0]);

  if (ctx.useCanvas) { return contentGroup.selectAll('*').remove(); }

  renderAxis(ctx, contentGroup, c);
  renderPointsSvg(ctx, contentGroup, c);
}

function renderAxis(ctx, group, c) {
  let axisGroup = cc.materialize(group, 'g', 'ssg-chart-axis', [0])
    .lower();

  cc.materialize(axisGroup, 'line', 'ssg-axis-x', [0])
    .attr('x1', ctx.xAxis).attr('x2', ctx.x1Chart)
    .attr('y1', ctx.yAxis).attr('y2', ctx.yAxis);

  cc.materialize(axisGroup, 'line', 'ssg-axis-y', [0])
    .attr('x1', ctx.xAxis).attr('x2', ctx.xAxis)
    .attr('y1', ctx.yAxis).attr('y2', ctx.y1Chart);
}

function renderPointsSvg(ctx, group, c) {
  const filteredLines = ctx.filtered.records;
  const filteredOutLines = ctx.showFilteredOutRecords ? filteredLines[1] : [];

  const x = cx(ctx, c);
  const y = cy(ctx, c);


  const outGroup = cc.materialize(group, 'g', 'ssg-filtered-out', [0]);

  cc.materialize(outGroup, 'circle', '', filteredOutLines)
    .attr('cx', x)
    .attr('cy', y)
    .attr('r', pointsRadius);


  const inGroup = cc.materialize(group, 'g', 'ssg-filtered-in', [0]);

  cc.materialize(inGroup, 'circle', '', filteredLines[0])
    .attr('cx', x)
    .attr('cy', y)
    .attr('r', pointsRadius)
    .attr('fill', line => ctx.scales.color(line[line.length - 1]));
}

function renderContentGroupCanvas(ctx, canvas, c) {
  const step = ctx.xStep;
  canvas.setTransform(1, 0, 0, 1,
    Math.round(c * step - ctx.viewBox[0]),
    Math.round(-ctx.viewBox[1]));

  renderAxisCanvas(ctx, canvas, c);
  renderPointsCanvas(ctx, canvas, c);
}

function renderAxisCanvas(ctx, canvas, c) {
  canvas.strokeStyle = '#bbb';
  canvas.beginPath();

  canvas.moveTo(ctx.xAxis, ctx.yAxis);
  canvas.lineTo(ctx.x1Chart, ctx.yAxis);

  canvas.moveTo(ctx.xAxis, ctx.yAxis);
  canvas.lineTo(ctx.xAxis, ctx.y1Chart);

  canvas.stroke();
}

function renderPointsCanvas(ctx, canvas, c) {
  const colorScale = ctx.scales.color;
  const PI2 = Math.PI * 2;

  const x = cx(ctx, c);
  const y = cy(ctx, c);

  const filteredLines = ctx.filtered.records;

  if (ctx.showFilteredOutRecords) {
    _.forEach(filteredLines[1], function renderPoint(line) {
      canvas.fillStyle = 'rgb(230,230,230)';
      canvas.beginPath();
      canvas.arc(x(line), y(line), pointsRadius, 0, PI2);
      canvas.fill();
    });
  }

  _.forEach(filteredLines[0], function renderPoint(line) {
    canvas.fillStyle = colorScale(line[line.length - 1]);
    canvas.beginPath();
    canvas.arc(x(line), y(line), pointsRadius, 0, PI2);
    canvas.fill();
  });
}


function renderOverlayGroup(ctx, group, c) {
  let overlayGroup = cc.materialize(group, 'g', 'ssg-overlay', [0]);

  renderExpand(ctx, overlayGroup, c);
  renderFilter(ctx, overlayGroup, c);
  renderHover(ctx, overlayGroup, c);
  renderLabels(ctx, overlayGroup, c);
}

function renderExpand(ctx, group, c) {
  cc.materialize(group, 'rect', 'ssg-overlay-bounds', [0])
    .attr('x', 0).attr('y', 0)
    .attr('width', ctx.xStep).attr('height', ctx.refSizeY);

  cc.materialize(group, 'text', 'ssg-expand-icon', [c])
    .attr('x', ctx.xAxis - axisLabelPadX)
    .attr('y', ctx.yAxis + axisLabelPadY)
    .attr('dy', '1em')
    .text('\uf065')
    .on('mousedown', ctx.scatterplotExpand);
}

function renderFilter(ctx, group, c) {
  let filterRect = ctx.filterRect;
  let mousePress = ctx.mousePress;

  let rect = cc.materialize(group, 'rect', 'ssg-filter', [0]);
  if (!filterRect || !mousePress || c !== mousePress.c) {
    return rect.remove();
  }

  rect.attr('x', filterRect.x0).attr('width', filterRect.x1 - filterRect.x0)
    .attr('y', filterRect.y0).attr('height', filterRect.y1 - filterRect.y0);
}

function renderHover(ctx, group, c) {
  const color = ctx.colors.highlight || 'black';

  const x = cx(ctx, c);
  const y = cy(ctx, c);

  const hoverGroup = cc.materialize(group, 'g', 'ssg-hover-group', [0]);
  const hoverLine = ctx.view.hover.line ? [ctx.view.hover.line] : [];

  cc.materialize(hoverGroup, 'circle', 'ssg-hover-point', hoverLine)
    .attr('cx', x)
    .attr('cy', y)
    .attr('r', pointsHoverRadius)
    .attr('fill', color);

  cc.materialize(hoverGroup, 'circle', 'ssg-hover-halo', hoverLine)
    .attr('cx', x)
    .attr('cy', y)
    .attr('r', pointsHaloRadius)
    .attr('stroke', color);
}

function renderLabels(ctx, group, c) {
  cc.materialize(group, 'text', 'ssg-axis-x-label', [0])
    .attr('x', ctx.xAxis + axisLabelPadX)
    .attr('y', ctx.yAxis + fontSize + axisLabelPadY)
    .attr('font-size', fontSize)
    .text(ctx.columns[c + 1].label);

  cc.materialize(group, 'text', 'ssg-axis-y-label', [0])
    .attr('x', -ctx.yAxis + axisLabelPadX)
    .attr('y', ctx.xAxis - axisLabelPadY - 2)
    .attr('font-size', fontSize)
    .attr('transform', 'rotate(-90 0,0)')
    .text(ctx.columns[c].label);
}


// State

function buildSizes(ctx) {
  const padding = {
    left: cc.padLeft, right: cc.padRight,
    top: 0, bottom: 0
  };

  const result = cc.buildStripSizes(ctx, padding);
  result.xStep = cc.translateStep(result.refSizeX, ctx.columns.length);

  result.xAxis = padChartX0 + .5;
  result.yAxis = Math.round(result.refSizeY - padChartY0 - fontSize) + .5;

  result.x0Chart = result.xAxis + axisPad;
  result.x1Chart = Math.round(result.xStep - padChartX1) + .5;

  result.y0Chart = result.yAxis - axisPad;
  result.y1Chart = padChartY1 + .5;

  result.chartRect = [
    result.x0Chart, result.y1Chart, result.x1Chart, result.y0Chart
  ];

  return result;
}

function buildScales(ctx) {
  const result = {};

  const x0 = ctx.x0Chart;
  const y0 = ctx.y0Chart;
  const xLen = ctx.x1Chart - ctx.x0Chart;
  const yLen = ctx.y1Chart - ctx.y0Chart;

  result.xScales = _.map(ctx.scales.columns.slice(1, ctx.scales.columns.length),
    scale => val => x0 + xLen * scale(val));

  result.yScales = _.map(ctx.scales.columns.slice(0, ctx.scales.columns.length - 1),
    scale => val => y0 + yLen * scale(val));

  return result;
}

function initState(base, ctx) {
  ctx = _.assign({
    viewChanges: {},

    mousePress: null,
    filterRect: null
  }, base, ctx);

  _.assign(ctx, buildSizes(ctx));
  _.assign(ctx, buildScales(ctx));

  return ctx;
}


// Exports

export default class ScatterStrip {
  constructor(opts) {
    this.baseCtx = opts;
    this.ctx = null;
  }

  render(ctx) {
    this.ctx = ctx = initState(this.baseCtx, ctx);

    d3.select(ctx.svg)
      .attr('width', ctx.sizeX)
      .attr('height', ctx.sizeY)
      .attr('viewBox', ctx.viewBox.join(' '));

    ctx.canvas.width = ctx.sizeX;
    ctx.canvas.height = ctx.sizeY;
    ctx.canvas.style.width = ctx.sizeX + 'px';
    ctx.canvas.style.height = ctx.sizeY + 'px';

    renderChartsStrip(ctx);

    initMouse(ctx);
  }

  refresh(viewChanges) {
    if (!this.ctx) { return; }

    let refreshContent = false;
    let refreshOverlay = false;

    _.chain(viewChanges).keys().forEach(viewData => {
      switch (viewData) {
        case 'colFilters': refreshContent = true;
        case 'hover': refreshOverlay = true;
      }
    }).commit();

    renderChartsStrip(this.ctx, refreshContent, refreshOverlay);
  }

  reset() {
    if (!this.ctx) { return; }

    d3.select(this.ctx.svg).selectAll('*').remove();
    this.ctx = null;
  }
};

