
import ParallelChart from './js/parallel_chart';
import ScatterStrip from './js/scatterstrip';
import ScatterPlot from './js/scatterplot';

import RenderData, * as renderdata from './js/renderdata';

import './parallel_vis.less';
import './styles/layout.css';
import './styles/utils.css';

import { uiModules } from 'ui/modules';

import _ from 'lodash';


const module = uiModules.get('kibana/parallel_lines_chart_vis', ['kibana']);


function refreshViewMode($scope) {
  if (!$scope.renderData.records.length) {
    $scope.viewMode = 'nodata';
  } else if ($scope.scatterplotAxes) {
    $scope.viewMode = 'bigsp';
  } else {
    $scope.viewMode = 'main';
  }
}

function render($scope) {
  let mainChartsWidth;
  let toReset;

  const { renderData } = $scope;

  switch ($scope.viewMode) {
    case 'main':
      mainChartsWidth = $scope.refs.chartsArea.getBoundingClientRect().width;

      $scope.parallelChart.render(_.assign({
        sizeX: mainChartsWidth,
        sizeY: $scope.refs.parallelChart.getBoundingClientRect().height
      }, renderData));

      $scope.scatterStrip.render(_.assign({
        sizeX: mainChartsWidth,
        sizeY: $scope.refs.scatterStrip.getBoundingClientRect().height
      }, renderData));

      toReset = ['scatterPlot'];

      break;

    case 'bigsp':
      $scope.scatterPlot.render(_.assign({
        scatterplotAxes: $scope.scatterplotAxes,
        pointsRadius: renderData.bigSpPointsRadius
      }, renderData));

      // Keeping main charts
      toReset = [];

      break;

    default:
      toReset = ['parallelChart', 'scatterStrip', 'scatterPlot'];
  }

  _($scope).pick(toReset).forEach(chart => chart.reset());

  return $scope;
}

function resetCharts($scope) {
  _($scope).pick(['parallelChart', 'scatterStrip', 'scatterPlot'])
    .forEach(chart => chart.reset());
}

function restoreColumnsVisibility() {
  this.renderData.all.columns.forEach(col => col.show = true);
  this.renderData.all.columns =
    renderdata.columnsFromVis(this.vis, this.renderData.all.columns);

  this.renderData.applyVisibility();
  this.editableParams.columns = this.renderData.all.columns;

  render(this);
}

function onViewChanges(viewChanges) {
  if (viewChanges.colFilters) {
    this.renderData.applyColumnFilters();
  }

  switch (this.viewMode) {
    case 'main':
      this.parallelChart.refresh(viewChanges);
      this.scatterStrip.refresh(viewChanges);
      break;

    case 'bigsp':
      this.scatterPlot.refresh(viewChanges);
      break;
  }
}

function onMoveColumn(srcColIdx, dstColIdx) {
  this.renderData.moveColumn(srcColIdx, dstColIdx);
  this.editableParams.columns = this.renderData.all.columns;

  render(this);
}

function onSortColumn(colIdx) {
  this.renderData.sortColumn(colIdx);
  this.editableParams.columns = this.renderData.all.columns;

  render(this);
}

function onHideColumn(colIdx) {
  if (this.renderData.columns.length <= 2) { return; }

  this.renderData.toggleColumnVisible(colIdx);
  this.editableParams.columns = this.renderData.all.columns;

  render(this);
}

function onScatterplotExpand(axes) {
  this.scatterplotAxes = axes;
  refreshViewMode(this);                                // size watching invokes render
}

function onScatterplotClose() {
  this.scatterplotAxes = null;
  refreshViewMode(this);
}


module.controller('ParallelLinesChartVisController', function (
  Private, $scope, $element
) {
  const { vis } = $scope;

  const element = $element[0];
  const editableVis = vis.getEditableVis();


  // Initialize state

  _.assign($scope, {
    editableParams: (editableVis || vis).params,
    renderData: new RenderData(),

    refs: {
      vis: element,
      scrollable: element.getElementsByClassName('parallel-scrollable')[0],
      chartsArea: element.getElementsByClassName('parallel-charts-area')[0],
      parallelChart: element.getElementsByClassName('parallel-chart')[0],
      scatterStrip: element.getElementsByClassName('parallel-scatter-strip')[0],
      scatterplot: element.getElementsByClassName('parallel-big-scatterplot')[0],

      parallelChartSvg:
        element.getElementsByClassName('parallel-chart-svg')[0],
      parallelChartCanvas:
        element.getElementsByClassName('parallel-chart-canvas')[0],

      scatterStripSvg:
        element.getElementsByClassName('parallel-scatter-strip-svg')[0],
      scatterStripCanvas:
        element.getElementsByClassName('parallel-scatter-strip-canvas')[0],

      scatterplotSvg:
        element.getElementsByClassName('parallel-big-scatterplot-svg')[0],
    },

    viewMode: 'nodata',
    scatterplotAxes: null,

    restoreColumnsVisibility
  });

  const { refs, editableParams } = $scope;

  const chartCallbacks = {
    onViewChanges: onViewChanges.bind($scope),
    onMoveColumn: onMoveColumn.bind($scope),
    onSortColumn: onSortColumn.bind($scope),
    onHideColumn: onHideColumn.bind($scope),
    onScatterplotExpand: onScatterplotExpand.bind($scope),
    onScatterplotClose: onScatterplotClose.bind($scope)
  };

  _.assign($scope, {
    parallelChart: new ParallelChart(_.assign({
      svg: refs.parallelChartSvg,
      canvas: refs.parallelChartCanvas
    }, chartCallbacks)),

    scatterStrip: new ScatterStrip(_.assign({
      svg: refs.scatterStripSvg,
      canvas: refs.scatterStripCanvas
    }, chartCallbacks)),

    scatterPlot: new ScatterPlot(_.assign({
      container: refs.scatterplot,
      svg: refs.scatterplotSvg
    }, chartCallbacks))
  });


  // Watches

  $scope.$watchCollection(function chartModeAndSizes(scope) {
    return [
      scope.viewMode,
      scope.refs.vis.clientWidth,
      scope.refs.vis.clientHeight,
      scope.refs.chartsArea.clientWidth,
      scope.refs.chartsArea.clientHeight
    ];
  }, _.debounce(function onResize() {
    render($scope);
  }, 50));

  if (editableVis) {
    $scope.$watch(
      scope => editableVis.getEnabledState(),
      _.debounce(function refreshColumns() {
        const { editableParams } = $scope;

        editableParams.columns =
          renderdata.columnsFromVis(editableVis, editableParams.columns);
      }, 50),
      true);
  }

  $scope.$on('$destroy', function onDestroy() {
    resetCharts($scope);                    // Removes eventual tooltips from body
  });

  $scope.$watch('esResponse', function buildRecordsAndRedraw(resp) {
    const { vis, editableParams } = $scope;
    const { aggs } = vis;

    const columns = renderdata.columnsFromVis(vis, editableParams.columns);
    const records = renderdata.recordsFromResponse(columns, aggs, resp);

    $scope.renderData = new RenderData(
      _.assign({}, editableParams, { columns, records }));

    refreshViewMode($scope);
    render($scope);
  });
});

module.controller('ParallelLinesChartVisColumnsController', function (
  Private, $scope, $element
) {
  $scope.hasColumnParameters = _([
    'number', 'integer', 'number_range', 'integer_range'
  ])
  .indexBy()
  .mapValues(_.constant(true))
  .value();
});
